from collections.abc import Iterator
from typing import Optional

from bs4 import BeautifulSoup, NavigableString, PageElement, ResultSet, Tag
from casetransform import transform

from .dissect import capture_values


def from_meta(soup: BeautifulSoup, prop: str) -> Optional[str]:
    tag = soup.find("meta", property=prop)
    return tag["content"] if tag.has_attr("content") else None


def long_title(soup: BeautifulSoup) -> Optional[Tag]:
    """The full title is used as (a) an anchor to get the ponente tag and (b) the source of the long title text

    Args:
        soup (BeautifulSoup): The source html document

    Returns:
        Optional[Tag]: The long title tag <h3> tag, if found
    """
    header = soup.find("center", class_="caseheader")
    return header.find("h3")


def get_headers(soup: BeautifulSoup) -> Optional[ResultSet]:
    """Used for picking items from the headers div

    Args:
        soup (BeautifulSoup): The source html document

    Returns:
        Optional[ResultSet]: h2 elements
    """
    class_list = "container case_document alternativeHeader"
    if not (header_block := soup.find("div", class_=class_list)):
        return None
    if not (header_elements := header_block("h2")):
        return None
    return header_elements


def pick(slice: int, headers: ResultSet) -> Optional[Tag]:
    try:
        return headers[slice]
    except IndexError:
        return None
    except TypeError:
        return None


def get_citation(soup: BeautifulSoup, idx: str) -> Optional[str]:
    attrs = {"data-docid": idx}
    if not (tag := soup.find("article", attrs=attrs)):
        return None
    if not isinstance(tag.next_element, NavigableString):
        return None
    return str(tag.next_element).strip()


def get_ponente_tag(tag: Tag) -> Tag:
    while True:
        if isinstance(tag, Tag) and tag.name == "strong":
            return tag
        tag = tag.next_element


def get_ponente_text(tag: Tag) -> Optional[str]:
    if not (ponente_tag := get_ponente_tag(tag)):
        return None
    return ponente_tag.get_text().title()


def index_from(
    soup: BeautifulSoup,
    target_tag: Tag,
    exclude_target: bool = True,
) -> int:
    index = str(soup).find(str(target_tag))
    if exclude_target:
        return index
    return index + len(str(target_tag))


def get_ponencia(soup: BeautifulSoup):
    text = str(soup)
    start = soup.find("div", attrs={"align": "justify"})
    end = soup.find("hr")
    if not start or not end:
        return None
    s = index_from(soup, start)
    e = index_from(soup, end)
    return text[s:e]


def get_post_ponencia(
    soup: BeautifulSoup,
    ponencia_text: str,
) -> Optional[str]:
    source = str(soup)
    terminal = source.find(ponencia_text) + len(ponencia_text)
    if not (end_tag := soup.find("doctags")):
        return None
    return source[terminal : source.find(str(end_tag))]


def get_annex(text: str) -> str:
    soup = BeautifulSoup(text, "html5lib")
    # no dividers found
    if not soup("hr"):
        return text

    # decompose first divider which marks the end of the ponencia
    soup("hr")[0].decompose()

    # revise the soup object string
    new_soup_text = str(soup)

    # no dividers found
    if not soup("hr"):
        return new_soup_text

    # use the first divider found as the terminal string
    target = soup("hr")[0]
    index = new_soup_text.find(str(target))
    new_slice = new_soup_text[:index]
    return new_slice


def cleaned_annex(
    soup: BeautifulSoup,
    ponencia_text: str,
    ponencia_footnotes: list,
) -> Optional[str]:
    # since no footnotes found in ponencia, no need to clean
    if not ponencia_footnotes:
        return None

    # get initial slice of annex after the ponencia
    raw = get_post_ponencia(soup, ponencia_text)
    if not raw:
        return None

    # get initial slice of annex after the ponencia
    annex = get_annex(raw)

    cleaned_annex = with_unique_fn(annex)
    if not cleaned_annex:
        return None

    return cleaned_annex


def fn_digits(matches: ResultSet) -> Iterator:
    for match in matches:
        content = match.get_text().strip("[] ")
        if content.isdigit():
            yield int(content)


def get_footnotes(text: Optional[str]):
    if not text:
        return None
    soup = BeautifulSoup(text, "html5lib")
    matches = soup("sup", class_="fn")
    if not matches:
        return None
    return list(fn_digits(matches))


def get_body(soup: BeautifulSoup):
    return str(soup.find("div", class_="container case_document"))


def with_unique_fn(text: str) -> Optional[str]:
    soup = BeautifulSoup(text, "html5lib")
    if not (footnotes_found := soup("sup")):
        return None
    for count, note in enumerate(footnotes_found, start=1):
        note["id"] = f"{count}"
    return str(soup)


def metadata(soup: BeautifulSoup, idx: str) -> dict:
    h = get_headers(soup)
    return {
        "title": from_meta(soup, "og:title"),
        "long_title": lt.get_text() if (lt := long_title(soup)) else None,
        "docket": from_meta(soup, "og:description"),
        "initial": get_citation(soup, idx),
        "composition": com.get_text().title() if (com := pick(0, h)) else None,
        "category": cat.get_text().title() if (cat := pick(3, h)) else None,
        "body": get_body(soup),
        "ponente": get_ponente_text(lt) if lt else None,
        "ponencia": (ponencia := get_ponencia(soup)),
        "digits": (notes := get_footnotes(ponencia) if ponencia else None),
        "annex": cleaned_annex(soup, ponencia, notes) if ponencia else None,
    }


def content_proper(text: str) -> dict:
    """Use the ruling phrase library to to be added to the calling function

    Args:
        text (str): The text of the ponencia

    Returns:
        dict: A dictionary with keys acceptable to scrapy processing pipeline.
    """
    keys = ["fallo", "voting", "ruling", "ruling_marker", "ruling_offset"]
    proper = capture_values(text)
    return {key: value for key, value in proper.items() if key in keys}


def initial_data(text: str, idx: str) -> dict:
    """Use the supplied parameters to generate initial data dictionary

    Args:
        text (str): The unprocessed string representing the ponencia
        idx (str): The url identifier

    Returns:
        dict: A dictionary with keys acceptable to scrapy processing pipeline.
    """
    soup = BeautifulSoup(text, "html5lib")
    h = get_headers(soup)
    ponencia = get_ponencia(soup)
    notes = get_footnotes(ponencia) if ponencia else None
    return {
        "initial": get_citation(soup, idx),
        "composition": com.get_text().title() if (com := pick(0, h)) else None,
        "category": cat.get_text().title() if (cat := pick(3, h)) else None,
        "ponencia": ponencia,
        "annex": cleaned_annex(soup, ponencia, notes) if ponencia else None,
    }


def legacy_data(text: str, idx: str) -> dict:
    """Use the supplied parameters to combine data dictionaries.

    Args:
        text (str): The unprocessed string representing the ponencia
        idx (str): The url identifier

    Returns:
        dict: A dictionary with keys acceptable to scrapy processing pipeline..
    """
    case_details = initial_data(text, idx)
    ponencia_details = content_proper(case_details["ponencia"])
    data_complete = case_details | ponencia_details
    data_complete["ponencia"] = with_unique_fn(data_complete["ponencia"])
    data_complete["ponencia"] = transform(data_complete["ponencia"])
    return data_complete
